#include <unistd.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <sys/types.h>
#include <fuse.h>

#include <fcntl.h>
#include <pthread.h>
#include <libgen.h>

#include <tcutil.h>
#include <tcbdb.h>
#include <tchdb.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>

#include "lib_safe.h"
#include "lib_log.h"
#include "lib_cfg.h"
#include "retcodes.h"
#ifdef LZO
#include "lib_lzo.h"
#else
#include "lib_qlz.h"
#endif
#include "lib_tc.h"

extern char *logname;
extern char *function;
extern int debug;
extern int BLKSIZE;
extern int max_threads;
extern BLKDTA **tdta;

TCHDB *dbb = NULL;
TCHDB *dbu = NULL;
TCHDB *dbp = NULL;
TCBDB *dbl = NULL;              // Hardlink
TCHDB *dbs = NULL;              // Symlink
TCHDB *dbdta = NULL;
TCBDB *dbdirent = NULL;
TCMDB *dbcache;
TCMDB *dbdtaq;
TCMDB *blkcache;

int written = 0;
static pthread_mutex_t lock_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t thread_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t write_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t tiger_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t open_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t qdta_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t qempty_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t moddb_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t dta_mutex[15];


#ifdef i386
#define ITERATIONS 30
#else
#define ITERATIONS 500
#endif

typedef unsigned long long int word64;
typedef unsigned long word32;
typedef unsigned char byte;
void tiger(byte *, word64, word64 *);

u_int32_t db_flags, env_flags;

#define die_dberr(t,f...) { LFATAL(f); exit(EXIT_DBERR); }
#define die_dataerr(f...) { LFATAL(f); exit(EXIT_DATAERR); }
#define die_syserr() { LFATAL("Fatal system error : %s",strerror(errno)); exit(EXIT_SYSTEM); }

unsigned char *binhash(unsigned char *buf, int size)
{
    word64 res[3];
    unsigned char *thehash;
    tiger((unsigned char *) buf, size, res);
    thehash = s_malloc(3 * sizeof(word64));
    memcpy(thehash, &res, 3 * sizeof(word64));
    return (thehash);
}

#ifdef DEBUG
void loghash(char *msg, unsigned char *bhash)
{
    char *ascii_hash = NULL;
    int n;
    char *p1, *p2;

    for (n = 0; n < (3 * sizeof(word64)); n++) {
        p1 = as_sprintf("%02X", bhash[n]);
        if (n == 0) {
            ascii_hash = s_strdup(p1);
        } else {
            p2 = s_strdup(ascii_hash);
            free(ascii_hash);
            ascii_hash = as_sprintf("%s%s", p2, p1);
            free(p2);
        }
        free(p1);
    }
    LDEBUG("%s : %s", msg, ascii_hash);
    free(ascii_hash);
}
# else
void loghash(char *msg, unsigned char *bhash)
{
}
# endif


void log_fatal_hash(char *msg, unsigned char *bhash)
{
    char *ascii_hash = NULL;
    int n;
    char *p1, *p2;

    for (n = 0; n < (3 * sizeof(word64)); n++) {
        p1 = as_sprintf("%02X", bhash[n]);
        if (n == 0) {
            ascii_hash = s_strdup(p1);
        } else {
            p2 = s_strdup(ascii_hash);
            free(ascii_hash);
            ascii_hash = as_sprintf("%s%s", p2, p1);
            free(p2);
        }
        free(p1);
    }
    LFATAL("%s : %s", msg, ascii_hash);
    free(ascii_hash);
}

TCHDB *hashdb_open(char *dbpath, int cacherow,
                   unsigned long long bucketsize)
{
    TCHDB *hdb;
    int ecode;

    FUNC;
    hdb = tchdbnew();
    if (cacherow > 0) {
        tchdbsetcache(hdb, cacherow);
    }
    tchdbtune(hdb, bucketsize, 0, 0, HDBTLARGE);
    if (!tchdbopen(hdb, dbpath, HDBOWRITER | HDBOCREAT)) {
        ecode = tchdbecode(hdb);
        die_dberr("open error: %s", tchdberrmsg(ecode));
    }
    EFUNC;
    return hdb;
}

void tc_defrag()
{
    flush_dta_queue();
    if (!tchdboptimize(dbb, atol(config->fileblockbs), 0, 0, HDBTLARGE))
        LINFO("fileblock.tch not optimized");
    if (!tchdboptimize(dbu, atol(config->blockusagebs), 0, 0, HDBTLARGE))
        LINFO("blockusage.tch not optimized");
    if (!tchdboptimize(dbp, atol(config->metabs), 0, 0, HDBTLARGE))
        LINFO("metadata.tcb not optimized");
    if (!tchdboptimize(dbs, atol(config->symlinkbs), 0, 0, HDBTLARGE))
        LINFO("metadata.tcb not optimized");
    if (!tchdboptimize(dbdta, atol(config->blockdatabs), 0, 0, HDBTLARGE))
        LINFO("blockdata.tch not optimized");
    if (!tcbdboptimize
        (dbdirent, 0, 0, atol(config->direntbs), -1, -1, BDBTLARGE))
        LINFO("dirent.tcb not optimized");
    if (!tcbdboptimize
        (dbl, 0, 0, atol(config->hardlinkbs), -1, -1, BDBTLARGE))
        LINFO("dirent.tcb not optimized");
}


void tc_open(bool defrag)
{
    char *dbpath;
    int ecode;

    FUNC;
    dbpath = as_sprintf("%s/fileblock.tch", config->fileblock);
    LDEBUG("Open database %s", dbpath);
    dbb = hashdb_open(dbpath, 2621440, atol(config->fileblockbs));
    free(dbpath);

    dbpath = as_sprintf("%s/blockusage.tch", config->blockusage);
    LDEBUG("Open database %s", dbpath);
    dbu = hashdb_open(dbpath, 2621440, atol(config->blockusagebs));
    free(dbpath);

    dbpath = as_sprintf("%s/metadata.tcb", config->meta);
    LDEBUG("Open database %s", dbpath);
    dbp = hashdb_open(dbpath, 1000, atol(config->metabs));
    free(dbpath);

    dbpath = as_sprintf("%s/blockdata.tch", config->blockdata);
    LDEBUG("Open database %s", dbpath);
    dbdta = hashdb_open(dbpath, 10, atol(config->blockdatabs));
    free(dbpath);

    dbpath = as_sprintf("%s/symlink.tch", config->symlink);
    LDEBUG("Open database %s", dbpath);
    dbs = hashdb_open(dbpath, 10, atol(config->symlinkbs));
    free(dbpath);

    dbdirent = tcbdbnew();
    tcbdbtune(dbdirent, 0, 0, atol(config->direntbs), -1, -1, BDBTLARGE);
    /* The dirent database is a B-TREE DB with cursors */
    dbpath = as_sprintf("%s/dirent.tcb", config->dirent);
    LDEBUG("Open database %s", dbpath);
    if (!tcbdbopen(dbdirent, dbpath, BDBOWRITER | BDBOCREAT)) {
        ecode = tcbdbecode(dbdirent);
        die_dberr("open error: %s", tcbdberrmsg(ecode));
    }
    free(dbpath);

    dbl = tcbdbnew();
    tcbdbtune(dbl, 0, 0, atol(config->hardlinkbs), -1, -1, BDBTLARGE);
    /* The dbl database is a B-TREE DB with cursors */
    dbpath = as_sprintf("%s/hardlink.tcb", config->hardlink);
    LDEBUG("Open database %s", dbpath);
    if (!tcbdbopen(dbl, dbpath, BDBOWRITER | BDBOCREAT)) {
        ecode = tcbdbecode(dbdirent);
        die_dberr("open error: %s", tcbdberrmsg(ecode));
    }
    free(dbpath);

    if (!defrag) {
        dbcache = tcmdbnew();
        dbdtaq = tcmdbnew();
        blkcache = tcmdbnew();
    }
    LDEBUG("All databases are open");
    EFUNC;
}

void hashdb_close(TCHDB * hdb)
{
    int ecode;

    FUNC;
    /* close the database */
    if (!tchdbclose(hdb)) {
        ecode = tchdbecode(hdb);
        die_dberr("close error: %s", tchdberrmsg(ecode));
    }
    /* delete the object */
    tchdbdel(hdb);
}

void tc_close(bool defrag)
{
    int ecode;

    FUNC;
    hashdb_close(dbb);
    hashdb_close(dbp);
    hashdb_close(dbu);
    hashdb_close(dbs);
    hashdb_close(dbdta);
    /* close the B-TREE database */
    if (!tcbdbclose(dbdirent)) {
        ecode = tcbdbecode(dbdirent);
        die_dberr("close error: %s", tchdberrmsg(ecode));
    }
    /* delete the object */
    tcbdbdel(dbdirent);

    /* delete the object */
    tcbdbdel(dbl);
    if (!defrag) {
        tcmdbdel(dbcache);
        tcmdbdel(dbdtaq);
        tcmdbdel(blkcache);
    }
    EFUNC;
}

void getLock()
{
    FUNC;
    pthread_mutex_lock(&lock_mutex);
    EFUNC;
    return;
}

void get_qdta_lock()
{
    FUNC;
    pthread_mutex_lock(&qdta_mutex);
    EFUNC;
    return;
}

void get_qempty_lock()
{
    FUNC;
    pthread_mutex_lock(&qempty_mutex);
    EFUNC;
    return;
}


void get_moddb_lock()
{
    FUNC;
    pthread_mutex_lock(&moddb_mutex);
    EFUNC;
    return;
}

void thread_lock()
{
    FUNC;
    pthread_mutex_lock(&thread_mutex);
    EFUNC;
    return;
}

void write_lock()
{
    FUNC;
    pthread_mutex_lock(&write_mutex);
    EFUNC;
    return;
}

void open_lock()
{
    FUNC;
    pthread_mutex_lock(&open_mutex);
    EFUNC;
    return;
}

void tiger_lock()
{
    FUNC;
    pthread_mutex_lock(&tiger_mutex);
    EFUNC;
    return;
}

void dta_lock(int count)
{
    pthread_mutex_lock(&dta_mutex[count]);
    EFUNC;
    return;
}

void all_dta_lock()
{
    int i;
    FUNC;
    for (i = 0; i < max_threads; i++) {
        pthread_mutex_lock(&dta_mutex[i]);
    }
    EFUNC;
    return;
}

void release_write_lock()
{
    pthread_mutex_unlock(&write_mutex);
    return;
}

void release_moddb_lock()
{
    pthread_mutex_unlock(&moddb_mutex);
    return;
}

void release_open_lock()
{
    pthread_mutex_unlock(&open_mutex);
    return;
}

void release_tiger_lock()
{
    pthread_mutex_unlock(&tiger_mutex);
    return;
}

void release_dta_lock(int count)
{
    FUNC;
    pthread_mutex_unlock(&dta_mutex[count]);
    EFUNC;
    return;
}

void release_all_dta_lock()
{
    int i;
    FUNC;
    for (i = 0; i < max_threads; i++) {
        pthread_mutex_unlock(&dta_mutex[i]);
    }
    EFUNC;
    return;
}

void dta_mutex_init(int count)
{
    int res;
    FUNC;
    res = pthread_mutex_init(&dta_mutex[count], NULL);
    if (res != 0)
        die_syserr();
    EFUNC;
    return;
}

void releaseLock()
{
    FUNC;
    pthread_mutex_unlock(&lock_mutex);
    return;
}

void release_qdta_lock()
{
    FUNC;
    pthread_mutex_unlock(&qdta_mutex);
    EFUNC;
    return;
}

void release_qempty_lock()
{
    FUNC;
    pthread_mutex_unlock(&qempty_mutex);
    EFUNC;
    return;
}

int thread_try_lock()
{
    int res;
    res = pthread_mutex_trylock(&lock_mutex);
    return (res);
}

int try_qempty_lock()
{
    int res;
    FUNC;
    res = pthread_mutex_trylock(&qempty_mutex);
    EFUNC;
    return(res);
}


void thread_lock_release()
{
    pthread_mutex_unlock(&thread_mutex);
    return;
}


char *s_dirname(char *path)
{
    char *tmpstr;
    char *rdir;
    char *retstr;

    if (NULL == path)
        return NULL;
    tmpstr = s_strdup(path);
    rdir = dirname(tmpstr);
    retstr = s_strdup(rdir);
    free(tmpstr);
    return retstr;
}

char *s_basename(char *path)
{
    char *tmpstr;
    char *rdir;
    char *retstr;
    tmpstr = s_strdup(path);
    rdir = basename(tmpstr);
    retstr = s_strdup(rdir);
    free(tmpstr);
    return retstr;
}

DBT *create_ddbuf(struct stat stbuf, char *filename)
{
    DBT *ddbuf;
    int len;

    FUNC;
    if (NULL != filename) {
        len = sizeof(struct stat) + strlen((char *) filename) + 1;
    } else
        len = sizeof(struct stat) + 1;

    ddbuf = s_malloc(sizeof(DBT));
    ddbuf->size = len;
    ddbuf->data = s_malloc(ddbuf->size);
    memcpy(ddbuf->data, &stbuf, sizeof(struct stat));
    if (NULL != filename) {
        memcpy(ddbuf->data + sizeof(struct stat), (char *) filename,
               strlen((char *) filename) + 1);
    } else
        memset(ddbuf->data + sizeof(struct stat), 0, 1);
    EFUNC;
    return ddbuf;
}

DBT *create_mem_ddbuf(MEMDDSTAT * ddstat)
{
    DBT *ddbuf;

    ddbuf = s_malloc(sizeof(DBT));
    ddbuf->data = s_malloc(sizeof(MEMDDSTAT));
    ddbuf->size = sizeof(MEMDDSTAT);
    memcpy(ddbuf->data, ddstat, sizeof(MEMDDSTAT));
    return ddbuf;
}

DDSTAT *value_to_ddstat(char *value, int size)
{
    DDSTAT *ddbuf;
    int filelen;

    FUNC;
    filelen = size - sizeof(struct stat);
    ddbuf = s_malloc(sizeof(DDSTAT));
    memcpy(&ddbuf->stbuf, value, sizeof(struct stat));
    if (1 == filelen) {
        memset(&ddbuf->filename, 0, MAX_POSIX_FILENAME_LEN);
    } else {
        memcpy(ddbuf->filename, value + sizeof(struct stat), filelen + 1);
    }
    LDEBUG("value_to_ddstat : return %llu", ddbuf->stbuf.st_ino);
    EFUNC;
    return ddbuf;
}

void dbmknod(const char *path, mode_t mode, char *linkdest, dev_t rdev)
{
    unsigned long long inode;

    FUNC;
    LDEBUG("dbmknod : %s", path);
    inode = get_next_inode();
    write_file_ent(path, inode, mode, linkdest, rdev);
    EFUNC;
    return;
}

void write_file_ent(const char *filename, unsigned long long inode,
                  mode_t mode, char *linkdest, dev_t rdev)
{
    struct stat stbuf;
    struct stat dirstat;
    time_t thetime;
    char *bname;
    char *parentdir;
    int res = 0;
    DBT *ddbuf;
    bool isdot = 0;
    bool isrootdir = 0;

    FUNC;
    LDEBUG("write_file_ent : filename %s, inodenumber %llu", filename,
           inode);
    bname = s_basename((char *) filename);
    parentdir = s_dirname((char *) filename);
    if (0 == strcmp(filename, "/"))
        isrootdir = 1;
    LDEBUG("write_file_ent: parentdir = %s", parentdir);
    write_nfi(inode + 1);

//Write stat structure to create an empty file.
    stbuf.st_ino = inode;
    stbuf.st_dev = 999988;
    stbuf.st_mode = mode;
    stbuf.st_nlink = 1;
    if (!isrootdir) {
        stbuf.st_uid = fuse_get_context()->uid;
        stbuf.st_gid = fuse_get_context()->gid;
    } else {
        stbuf.st_uid = 0;
        stbuf.st_gid = 0;
    }
    stbuf.st_rdev = rdev;
    if (S_ISDIR(mode)) {
        stbuf.st_size = 4096;
        stbuf.st_blocks = 1;
    } else {
        stbuf.st_size = 0;
        stbuf.st_blocks = 0;
    }
    if (S_ISLNK(mode)) {
        stbuf.st_size = 3;
        stbuf.st_blocks = 1;
    }
    stbuf.st_blksize = BLKSIZE;
    thetime = time(NULL);
    stbuf.st_atime = thetime;
    stbuf.st_mtime = thetime;
    stbuf.st_ctime = thetime;

    ddbuf = create_ddbuf(stbuf, bname);
    bin_write_dbdata(dbp, &inode, sizeof(inode), ddbuf->data, ddbuf->size);
    DBTfree(ddbuf);
    if (NULL != linkdest) {
        bin_write_dbdata(dbs, &inode, sizeof(unsigned long long), linkdest,
                         strlen(linkdest));
    }
    if (0 == strcmp(bname, "."))
        isdot = 1;
    if (0 == strcmp(bname, ".."))
        isdot = 1;
  recurse:
    if (S_ISDIR(mode) && isdot != 1) {
        btbin_write_dup(dbdirent, &stbuf.st_ino, sizeof(stbuf.st_ino),
                        &inode, sizeof(inode));
    } else {
        res = dbstat(parentdir, &dirstat);
        btbin_write_dup(dbdirent, &dirstat.st_ino, sizeof(dirstat.st_ino),
                        &inode, sizeof(inode));
    }
    if (S_ISDIR(mode) && !isdot && !isrootdir) {
        // Create the link inode to the previous directory
        LDEBUG("Create the link inode to the previous directory");
        isdot = 1;
        // Only if !isrootdir. Nothing lower the root.
        goto recurse;
    }
    free(parentdir);
    free(bname);
    EFUNC;
}

void write_nfi(unsigned long long nextinode)
{
    bin_write_dbdata(dbp, (unsigned char *) "NFI", strlen("NFI"),
                     (unsigned char *) &nextinode, sizeof(nextinode));
    return;
}

void formatfs()
{
    struct stat stbuf;
    unsigned long long nextinode = 1;
    FUNC;
    if (NULL == dbp) {
        tc_open(0);
    }
    stat(config->blockdata, &stbuf);
    write_nfi(nextinode);
    fs_mkdir("/", stbuf.st_mode);
    tc_close(0);
    return;
}

int get_dir_inode(char *dname, struct stat *stbuf)
{
    char *p;
    int depth = 0;
    DDSTAT *filestat = NULL;
    int res = 0;
    unsigned long long inode = 1;

    FUNC;
    while (1) {
        p = strchr(dname, '/');
        if (NULL == p)
            break;
        p[0] = 0;
        LDEBUG("p=%s", p);
        if (depth == 0) {
            LDEBUG("Lookup inode 1");
            filestat =
                dnode_bname_to_inode(&inode, sizeof(unsigned long long),
                                     "/");
            LDEBUG("got inode %llu", filestat->stbuf.st_ino);
        } else {
            ddstatfree(filestat);
            inode = stbuf->st_ino;
            filestat =
                dnode_bname_to_inode(&inode, sizeof(unsigned long long),
                                     dname);
            if (NULL == filestat) {
                res = -ENOENT;
                break;
            }
        }
        memcpy(stbuf, &filestat->stbuf, sizeof(struct stat));
        LDEBUG("After memcpy %llu", stbuf->st_ino);
        p++;
        depth++;
        dname = p;
        if (NULL == p)
            break;
    }
    if (res == 0) {
        ddstatfree(filestat);
        LDEBUG("return stbuf.st_ino=%llu", stbuf->st_ino);
    }
    EFUNC;
    return (res);
}

/* Fill struct stat from cache if present in the cache
   return 1 when found or 0 when not found in cache. */
int get_realsize_fromcache(unsigned long long inode,struct stat *stbuf)
{
    int result=0;
    DBT *data;
    MEMDDSTAT *mddstat;

    data = search_memhash(dbcache, &inode, sizeof(unsigned long long));
    if (data == NULL) {
        LDEBUG("inode %llu not found use size from database.", inode);
        return(result);
    }
    mddstat = value_tomem_ddstat((char *) data->data, data->size);
    memcpy(stbuf, &mddstat->stbuf, sizeof(struct stat));
    LDEBUG("get_realsize_fromcache : return stbuf from cache : size %llu",stbuf->st_size);
    DBTfree(data);
    memddstatfree(mddstat);
    return(result);
}

int dbstat(const char *filename, struct stat *stbuf)
{
    int retcode = 0;
    char *dname = NULL;
    char *bname = NULL;
    char *dupdname = NULL;
    char *mdupdname = NULL;
    DDSTAT *filestat = NULL;

    FUNC;
    dname = s_dirname((char *) filename);
    bname = s_basename((char *) filename);
    dupdname = s_strdup((char *) filename);
    mdupdname = dupdname;

// Walk the directory
    retcode = get_dir_inode(dupdname, stbuf);
    if (0 == retcode) {
        if (0 != strcmp(bname, dname)) {        /* This is the rootdir */
            // Now find the file within the directory
            filestat =
                dnode_bname_to_inode(&stbuf->st_ino,
                                     sizeof(unsigned long long), bname);
            if (NULL != filestat) {
                if (0 == strcmp(bname, filestat->filename)) {
                    if ( 0 == get_realsize_fromcache(filestat->stbuf.st_ino,stbuf)){
                        memcpy(stbuf, &filestat->stbuf, sizeof(struct stat));
                    }
                } else {
                    retcode = -ENOENT;
                }
                ddstatfree(filestat);
            } else {
                retcode = -ENOENT;
            }
        }
    } else
        retcode = -ENOENT;

    free(mdupdname);
    free(bname);
    free(dname);
    if (retcode == -ENOENT)
        LDEBUG("dbstat : File %s not found.", filename);
    EFUNC;
    return (retcode);
}


void ddstatfree(DDSTAT * ddstat)
{
    if (NULL != ddstat) {
        LDEBUG("ddstatfree really free");
        free(ddstat);
        ddstat = NULL;
    }
}

MEMDDSTAT *value_tomem_ddstat(char *value, int size)
{
    MEMDDSTAT *ddbuf;
    ddbuf = s_malloc(size);
    memcpy(ddbuf, value, size);
    return ddbuf;
}

void memddstatfree(MEMDDSTAT * ddstat)
{
    if (NULL != ddstat) {
        LDEBUG("memddstatfree really free");
        free(ddstat);
    }
    ddstat = NULL;
    return;
}

void comprfree(compr * compdata)
{
    if (compdata) {
        if (compdata->data)
            free(compdata->data);
        free(compdata);
    }
}

unsigned long long readBlock(unsigned long long blocknr,
                             const char *filename, char *blockdata,
                             unsigned long long inode)
{
    int ret;
    DBT *data;
    DBT *cachedata;
    unsigned char *tiger;
    unsigned char *dtiger;
    compr *uncompdata = NULL;
    INOBNO inobno;
    BLKCACHE *blk;

    FUNC;
    inobno.inode = inode;
    inobno.blocknr = blocknr;
    data = try_block_cache(inode, blocknr, 0);
    if (NULL != data) {
        LDEBUG("readBlock : block %llu - %llu found in cache", inode,
               blocknr);
        memcpy(blockdata, data->data, data->size);
        ret = data->size;
        DBTfree(data);
        return (ret);
    }
    data = check_block_exists(inobno);
    if (NULL == data) {
        return (0);
    }
// Not needed to copy this.
    tiger = s_malloc(data->size);
    memcpy(tiger, data->data, data->size);
    DBTfree(data);
// First try the cache
    get_moddb_lock();
    data = search_memhash(dbdtaq, tiger, TIGERLEN);
    release_moddb_lock();
    if ( NULL == data ) { 
       LDEBUG("readBlock hash not found in dbdtaq");
       data = search_dbdata(dbdta, tiger, TIGERLEN);
       if (NULL == data) {
           LDEBUG("readBlock hash not found in dbdta");
           cachedata = search_memhash(blkcache, &inobno.inode, sizeof(unsigned long long));
           if ( NULL != cachedata ) {
             blk = (BLKCACHE *) cachedata->data;
             dtiger=binhash(blk->blockdata, BLKSIZE);
             if ( 0 == memcmp(dtiger,tiger,TIGERLEN)){
                data=s_malloc(sizeof(DBT));
                data->data=s_malloc(BLKSIZE);
                data->size=BLKSIZE;
                memcpy(data->data,blk->blockdata,BLKSIZE);
                DBTfree(cachedata);
             } else {
                loghash("readBlock : unable to find n dbdta block hash :", tiger);
                die_dataerr("No data found to read.");
             }
             free(dtiger);
           } else {
              loghash("readBlock : unable to find n dbdta block hash :", tiger);
              die_dataerr("No data found to read.");
           }
       }
    }
    if (data->size != BLKSIZE) {
#ifdef LZO
        uncompdata = lzo_decompress(data->data, data->size);
#else
        uncompdata = clz_decompress(data->data, data->size);
#endif
        memcpy(blockdata, uncompdata->data, uncompdata->size);
        ret = uncompdata->size;
        comprfree(uncompdata);
    } else {
        memcpy(blockdata, data->data, data->size);
        ret = data->size;
    }
    DBTfree(data);
    free(tiger);
    EFUNC;
    return (ret);
}

/* Return the number of times this block is linked to files */
unsigned long long getInUse(unsigned char *tigerstr)
{
    unsigned long long counter;
    DBT *data;

    if (NULL == tigerstr)
        return (0);
    data = search_dbdata(dbu, tigerstr, TIGERLEN);
    if (NULL == data) {
        LDEBUG("getInuse nothing found return 0.");
        return (0);
    }
    memcpy(&counter, data->data, sizeof(counter));
    DBTfree(data);
    return counter;
}

void DBTfree(DBT * data)
{
    if (data) {
        if (data->data)
            free(data->data);
        free(data);
    }
}

void updateInUse(unsigned char *hashdata, int hashlen,
                 unsigned long long counter)
{
    bin_write_dbdata(dbu, hashdata, hashlen, (unsigned char *) &counter,
                     sizeof(counter));
    return;
}

void btbin_curwrite_dbdata(TCBDB * db, BDBCUR * cur, char *data,
                           int datalen)
{
    int ecode;
    FUNC;
    if (!tcbdbcurput(cur, data, datalen, BDBCPAFTER)) {
        ecode = tcbdbecode(db);
        die_dberr("tcbdbput2 failed : %s", tcbdberrmsg(ecode));
    }
    tcbdbsync(db);
    EFUNC;
}

void btasc_write_dbdata(TCBDB * db, char *keydata, char *dataData)
{
    int ecode;
    FUNC;
    if (!tcbdbputdup2(db, (char *) keydata, (char *) dataData)) {
        ecode = tcbdbecode(db);
        die_dberr("tcbdbput failed : %s", tcbdberrmsg(ecode));
    }
    tcbdbsync(db);
    EFUNC;
}

void btbin_write_dup(TCBDB * db, void *keydata, int keylen,
                     void *dataData, int datalen)
{
    int ecode;
    FUNC;
    if (!tcbdbputdup(db, keydata, keylen, dataData, datalen)) {
        ecode = tcbdbecode(db);
        die_dberr("tcbdbputdup failed : %s", tchdberrmsg(ecode));
    }
    tcbdbsync(db);
    EFUNC;
}

void btasc_curwrite_dbdata(TCBDB * db, BDBCUR * cur, unsigned char *data)
{
    int ecode;
    FUNC;

    if (!tcbdbcurput2(cur, (const char *) data, BDBCPAFTER)) {
        ecode = tcbdbecode(db);
        die_dberr("tcbdbput2 failed : %s", tcbdberrmsg(ecode));
    }
    tcbdbsync(db);
    EFUNC;
}

void btbin_write_dbdata(TCBDB * db, void *keydata, int keylen,
                        void *dataData, int datalen)
{
    int ecode;
    FUNC;
    if (!tcbdbput(db, keydata, keylen, dataData, datalen)) {
        ecode = tcbdbecode(db);
        die_dberr("tcbdbput failed : %s", tchdberrmsg(ecode));
    }
    tcbdbsync(db);
    EFUNC;
}

void mbin_write_dbdata(TCMDB * db, void *keydata, int keylen,
                       void *dataData, int datalen)
{
    FUNC;
    tcmdbput(db, keydata, keylen, dataData, datalen);
    EFUNC;
}

void bin_write_dbdata(TCHDB * db, void *keydata, int keylen,
                      void *dataData, int datalen)
{
    int ecode;
    if (!tchdbputasync(db, keydata, keylen, dataData, datalen)) {
        ecode = tchdbecode(db);
        die_dberr("tchdbputasync failed : %s", tchdberrmsg(ecode));
    }
}

void asc_write_dbdata(TCHDB * db, unsigned char *keydata,
                      unsigned char *dataData)
{
    int ecode;
    FUNC;
    if (!tchdbputasync
        (db, keydata, strlen((char *) keydata), dataData,
         strlen((char *) dataData))) {
        ecode = tchdbecode(db);
        die_dberr("tchdbputasync failed : %s", tchdberrmsg(ecode));
    }
    EFUNC;
}

/* Search in directory with inode dinode or key dinode for name bname 
   Return the inode of file bname                       */
DDSTAT *dnode_bname_to_inode(void *dinode, int dlen, char *bname)
{
    BDBCUR *cur;
    char *dbkey;
    int dbsize;
    char *dbvalue;
    DBT *statdata;
    DDSTAT *filestat = NULL;
    char *filename;
    DINOINO dinoino;
    unsigned long long keynode;
    unsigned long long valnode;

    FUNC;
    cur = tcbdbcurnew(dbdirent);
    if (!tcbdbcurjump(cur, (char *) dinode, dlen)
        && tcbdbecode(dbdirent) != TCESUCCESS) {
        tcbdbcurdel(cur);
        return NULL;
    }

    /* traverse records */
    while ((dbkey = tcbdbcurkey(cur, &dbsize)) != NULL) {
        if (0 != memcmp(dbkey, dinode, dlen)) {
            free(dbkey);
            break;
        }
        dbvalue = tcbdbcurval(cur, &dbsize);
        if (dbvalue) {
            memcpy(&valnode, dbvalue, sizeof(valnode));
            memcpy(&keynode, dbkey, sizeof(keynode));
            if (keynode == valnode && keynode != 1) {
                free(dbvalue);
                free(dbkey);
                tcbdbcurnext(cur);
                continue;
            }
            statdata =
                search_dbdata(dbp, &valnode, sizeof(unsigned long long));
            if ( NULL == statdata )
                die_dataerr
                ("Unable to find file existing in dbp, run fsck");
            filestat =
                value_to_ddstat((char *) statdata->data, statdata->size);
            DBTfree(statdata);
            free(dbvalue);
            if (0 != filestat->filename[0]) {
                LDEBUG("compare bname %s with filestat->filename %s",
                       bname, filestat->filename);
                if (0 == strcmp(bname, filestat->filename)) {
                    free(dbkey);
                    break;
                }
            } else {
                memcpy(&dinoino.dirnode, dinode,
                       sizeof(unsigned long long));
                dinoino.inode = filestat->stbuf.st_ino;
                filename =
                    btsearch_keyval(dbl, &dinoino, sizeof(DINOINO), bname,
                                    strlen(bname));
                if (NULL != filename) {
                    memcpy(&filestat->filename, filename,
                           strlen(filename) + 1);
                    free(filename);
                    free(dbkey);
                    break;
                }
            }
            ddstatfree(filestat);
            filestat = NULL;
        }
        free(dbkey);
        tcbdbcurnext(cur);
    }
    tcbdbcurdel(cur);
    if (NULL != filestat) {
        LDEBUG("dnode_bname_to_inode : filestat->filename=%s inode %llu",
               filestat->filename, filestat->stbuf.st_ino);
    } else {
        LDEBUG("dnode_bname_to_inode : return NULL");
    }
    return filestat;
}

DBT *search_dbdata(TCHDB * db, void *key, int len)
{
    DBT *data;
    int size;

    data = s_malloc(sizeof(DBT));
    data->data = tchdbget(db, key, len, &size);
    data->size = (unsigned long) size;
    if (NULL == data->data) {
        LDEBUG("search_dbdata : return NULL");
        free(data);
        data = NULL;
    } else
        LDEBUG("search_dbdata : return %lu bytes", data->size);
    return data;
}

unsigned long long get_next_inode()
{
    DBT *data;
    unsigned long long nextinode = 0;
    FUNC;

    data = search_dbdata(dbp, (unsigned char *) "NFI", strlen("NFI"));
    if (NULL != data) {
        memcpy(&nextinode, data->data, sizeof(nextinode));
        DBTfree(data);
    }
    LDEBUG("Found next inode number: %llu", nextinode);
    EFUNC;
    return (nextinode);
}

DBT *search_memhash(TCMDB * db, void *key, int len)
{
    DBT *data;
    int size;

    FUNC;
    data = s_malloc(sizeof(DBT));
    data->data = tcmdbget(db, key, len, &size);
    data->size = (unsigned long) size;
    if (NULL == data->data) {
        LDEBUG("search_memhash : return NULL");
        free(data);
        data = NULL;
    } else
        LDEBUG("search_memhash : return %lu bytes", data->size);
    EFUNC;
    return data;
}

void qdta(unsigned char *stiger, unsigned char *data, unsigned int size)
{
    get_moddb_lock();
       loghash("qdta",stiger);
       mbin_write_dbdata(dbdtaq, stiger, TIGERLEN,data,size);
    release_moddb_lock();
}

void addBlock(BLKDTA * blkdta)
{
    unsigned long long inuse;
    INOBNO inobno;
    DBT *cachedata=NULL;
    bool update=1;

    inobno.inode = blkdta->inode;
    inobno.blocknr = blkdta->blocknr;
    inuse = getInUse(blkdta->stiger);
    if (inuse == 0) {
        if ( blkdta->bsize + blkdta->offsetblock < BLKSIZE ) {
// Flush the blockcache before overwriting.
           cachedata=try_block_cache(blkdta->inode, blkdta->blocknr,0); 
           if ( cachedata ) DBTfree(cachedata);
           add_blk_to_cache(blkdta->inode, blkdta->blocknr, blkdta->blockfiller);
           update=0;
        }
        if (NULL == blkdta->compressed) {
#ifdef LZO
           blkdta->compressed =
               lzo_compress(blkdta->blockfiller, BLKSIZE);
#else
           blkdta->compressed =
               clz_compress(blkdta->blockfiller, BLKSIZE);
#endif
        }
        LDEBUG("Compressed %i bytes to %lu bytes", BLKSIZE,
               blkdta->compressed->size);
        if ( blkdta->bsize + blkdta->offsetblock == BLKSIZE ) {
             qdta(blkdta->stiger,blkdta->compressed->data,blkdta->compressed->size);
             loghash("addBlock queued with qdta",blkdta->stiger);
        }
        update_filesize(blkdta->inode, blkdta->bsize, blkdta->offsetblock,
                       blkdta->blocknr, blkdta->sparse,
                       blkdta->compressed->size, 0);
    } else {
        update_filesize(blkdta->inode, blkdta->bsize, blkdta->offsetblock,
                       blkdta->blocknr, blkdta->sparse, 0, 1);
    }
    if (NULL != blkdta->compressed)
        comprfree(blkdta->compressed);
    if ( update ) {
       inuse++;
       updateInUse(blkdta->stiger, TIGERLEN, inuse);
       bin_write_dbdata(dbb, &inobno, sizeof(INOBNO), blkdta->stiger,
                     TIGERLEN);
    }
    return;
}

void update_filesize_onclose(unsigned long long inode)
{
    DBT *dataptr;
    MEMDDSTAT *ddstat;
    dataptr = search_memhash(dbcache, &inode, sizeof(unsigned long long));
    if (dataptr == NULL) {
        LDEBUG("inode %llu not found to update.", inode);
        return;
    }
    ddstat = value_tomem_ddstat((char *) dataptr->data, dataptr->size);
    DBTfree(dataptr);
    flush_dta_queue();
    if (ddstat->updated > 0) {
        hash_update_filesize(ddstat, inode);
    } 
    memddstatfree(ddstat);
    return;
}

void update_filesize(unsigned long long inode, unsigned long long fsize,
                    unsigned int offsetblock, unsigned long long blocknr,
                    bool sparse,
                    unsigned int compressed, unsigned int deduplicated)
{
    DBT *dataptr;
    DBT *tigerdata;
    MEMDDSTAT *ddstat;
    DBT *ddbuf;
    int addblocks;
    INOBNO inobno;

    FUNC;

    LDEBUG
        ("update_filesize : inode %llu fsize %llu offset %u blocknet %llu bool %c",
         inode, fsize, offsetblock, blocknr, sparse);
    dataptr = search_memhash(dbcache, &inode, sizeof(unsigned long long));
    if (dataptr == NULL)
        return;
    ddstat = value_tomem_ddstat((char *) dataptr->data, dataptr->size);
    ddstat->updated++;
    ddstat->lzo_compressed_size = ddstat->lzo_compressed_size + compressed;
    ddstat->deduplicated = ddstat->deduplicated + deduplicated;
    ddstat->blocknr = blocknr;

    addblocks = fsize / 512;
    LDEBUG("update_filesize : addblocks = %i",addblocks);
    if ((ddstat->stbuf.st_blocks + addblocks) * 512 <
        ddstat->stbuf.st_size + fsize)
        addblocks++;
    // The file has not grown in size. This is an updated block.
    if (!sparse && ((blocknr * BLKSIZE) + offsetblock + fsize) <=
        ddstat->stbuf.st_size) {
        inobno.inode=inode;
        inobno.blocknr=blocknr;
        tigerdata=check_block_exists(inobno);
        if ( NULL != tigerdata ) {
           LDEBUG
               ("The file has not grown in size and the block exists. This is an updated block.");
           ddbuf = create_mem_ddbuf(ddstat);
           mbin_write_dbdata(dbcache, &inode, sizeof(unsigned long long),
                             (void *) ddbuf->data, ddbuf->size);
           DBTfree(ddbuf);
           DBTfree(dataptr);
           memddstatfree(ddstat);
           return;
        } else DBTfree(tigerdata);
    }
    // The file size has grown or the block is sparse.
    if (!sparse) {
        ddstat->stbuf.st_size = ddstat->stbuf.st_size + fsize;
        if (ddstat->stbuf.st_size > (512 * ddstat->stbuf.st_blocks)) {
            ddstat->stbuf.st_blocks = ddstat->stbuf.st_blocks + addblocks;
            LDEBUG("The file is not sparse and we need to add %i blocks",
                   addblocks);
        }
    } else {
        LDEBUG("The file adds a sparse block : add %i blocks", addblocks);
        ddstat->stbuf.st_blocks = ddstat->stbuf.st_blocks + addblocks;
    }
    ddbuf = create_mem_ddbuf(ddstat);
    mbin_write_dbdata(dbcache, &inode, sizeof(unsigned long long),
                      (void *) ddbuf->data, ddbuf->size);
    DBTfree(dataptr);
    DBTfree(ddbuf);
// Do not flush data until cachesize is reached
    if (ddstat->updated > config->cachesize ) {
        flush_dta_queue();
        hash_update_filesize(ddstat, inode);
        ddstat->updated = 0;
        ddbuf = create_mem_ddbuf(ddstat);
        mbin_write_dbdata(dbcache, &inode, sizeof(unsigned long long),
             (void *) ddbuf->data, ddbuf->size);
        DBTfree(ddbuf);
    }
    memddstatfree(ddstat);
    EFUNC;
    return;
}

void flush_dta_queue()
{
    FUNC;
    release_qdta_lock();
    get_qempty_lock();
    EFUNC;
}

void hash_update_filesize(MEMDDSTAT * ddstat, unsigned long long inode)
{
    DBT *ddbuf;
// Wait until the data queue is written before we update the filesize
    if (ddstat->stbuf.st_nlink > 1) {
        ddbuf = create_ddbuf(ddstat->stbuf, NULL);
    } else {
        ddbuf = create_ddbuf(ddstat->stbuf, ddstat->filename);
    }
    bin_write_dbdata(dbp, &inode,
                     sizeof(unsigned long long), (void *) ddbuf->data,
                     ddbuf->size);
    DBTfree(ddbuf);
    return;
}

/* delete = 1 Do delete dbdata
   delete = 0 Do not delete dbdta */
unsigned int commit_block(unsigned char *dbdata, unsigned char *chksum,
                  INOBNO inobno, bool delete)
{
    unsigned char *stiger;
    unsigned char *dtiger;
    compr *compressed;
    unsigned long long inuse;
    unsigned int res=0;
    BLKCACHE *blk;
    DBT *data;

    FUNC;
    stiger = (unsigned char *) binhash(dbdata, BLKSIZE);
#ifdef LZO
    compressed = lzo_compress((unsigned char *) dbdata, BLKSIZE);
#else
    compressed = clz_compress((unsigned char *) dbdata, BLKSIZE);
#endif
    res=compressed->size;
    // Now check if the block is used by others.
    if ( NULL != chksum ) {
       inuse = getInUse(chksum);
       if (inuse == 1) {
           loghash("commit_block : delete hash",chksum);
           delete_key(dbu, chksum, TIGERLEN);
           get_moddb_lock();
           if (!tchdbout(dbdta, chksum, TIGERLEN)) {
              loghash("commit_block : hash not found in dbdta, try cache", chksum);
              if (!tcmdbout(dbdtaq, chksum, TIGERLEN)){
                 data = search_memhash(blkcache, &inobno.inode, sizeof(unsigned long long));
                 if ( NULL != data ){
                   blk = (BLKCACHE *) data->data;
                   dtiger=binhash(blk->blockdata, BLKSIZE);
                   loghash("commit_block : dtiger deleted from cache",dtiger);
                   if ( 0 == memcmp(dtiger,stiger,TIGERLEN)){
                      mdelete_key(blkcache, &inobno.inode, sizeof(unsigned long long));
                      DBTfree(data);
                   } else die_dataerr("Weird scenes inside the coal mine.");
                   free(dtiger);
                 } else {
                     die_dataerr("Weird scenes inside the coal mine.");
                 }
              }
           }
           release_moddb_lock();
       } else {
           if ( inuse > 1 ) inuse--;
           updateInUse(chksum, TIGERLEN, inuse);
       }
    }
    inuse = getInUse(stiger);
    inuse++;
    updateInUse(stiger, TIGERLEN, inuse);
    qdta(stiger, (unsigned char *) compressed->data, compressed->size);
    comprfree(compressed);
    loghash("commit_block write hash :",stiger);
    bin_write_dbdata(dbb, &inobno, sizeof(INOBNO), stiger, TIGERLEN);
    free(stiger);
    return(res);
}

void add_blk_to_cache(unsigned long long inode, unsigned long long blocknr,
                      unsigned char *data)
{
    BLKCACHE blk;

    FUNC;
    blk.blocknr = blocknr;
    memcpy(&blk.blockdata, data, BLKSIZE);
    mbin_write_dbdata(blkcache, &inode, sizeof(unsigned long long),
                      (void *) &blk, sizeof(BLKCACHE));
    EFUNC;
    return;
}

/* mode = 0 -> update the block and delete dbdta
   mode = 1 -> flush the inode
   mode = 2 -> dbdta has not been written so don't delete
*/
DBT *try_block_cache(unsigned long long inode, unsigned long long blocknr,
                     unsigned int mode)
{
    DBT *retdata = NULL;
    DBT *data;
    DBT *tigerdata=NULL;
    INOBNO inobno;
    BLKCACHE *blk;

    FUNC;
    data = search_memhash(blkcache, &inode, sizeof(unsigned long long));
    if (data != NULL) {
        blk = (BLKCACHE *) data->data;
        if ((blocknr == blk->blocknr) && (mode != 1)) {
            retdata = s_malloc(sizeof(DBT));
            retdata->size = BLKSIZE;
            retdata->data = s_malloc(BLKSIZE);
            memcpy(retdata->data, blk->blockdata, BLKSIZE);
        } else {
            inobno.inode = inode;
            inobno.blocknr = blk->blocknr;
            tigerdata=check_block_exists(inobno);
            if ( NULL == tigerdata ) {
                mode=2;
            }
            if ( mode == 2 ) {
                commit_block(blk->blockdata, NULL, inobno, 0);
            } else {
                commit_block(blk->blockdata, tigerdata->data, inobno, 1);
            }
            if ( NULL != tigerdata ) DBTfree(tigerdata);
            mdelete_key(blkcache, &inode, sizeof(unsigned long long));
        }
        DBTfree(data);
    }
    return retdata;
}

void update_block(const char *blockdata, unsigned long long blocknr,
                  unsigned int offsetblock,
                  unsigned long long size, unsigned long long inode,
                  unsigned char *chksum)
{
    DBT *data;
    DBT *cachedata;
    unsigned char *dbdata;
    INOBNO inobno;
    compr *uncompdata;
    BLKCACHE *blk;
    unsigned char *dtiger;

    FUNC;
    LDEBUG("updateBlock : inode %llu blocknr %llu offsetblock %llu, size %llu",
           inode, blocknr, (unsigned long long) offsetblock,
           (unsigned long long) size);
    inobno.inode = inode;
    inobno.blocknr = blocknr;

    dbdata = s_malloc(BLKSIZE);
    memset(dbdata, 0, BLKSIZE);
    data = try_block_cache(inode, blocknr, 0);
    if (NULL != data) {
        memcpy(dbdata, data->data, data->size);
        memcpy(dbdata + offsetblock, blockdata, size);
        add_blk_to_cache(inode, blocknr, dbdata);
        update_filesize(inode, size, offsetblock, blocknr, 0, 0, 0);
        free(dbdata);
        DBTfree(data);
        return;
    } else
        LDEBUG("update_block: block not found in cache.");

// We don't need the old blockdata when we overwrite it completely anyway.

    if ( offsetblock > 0 || size < BLKSIZE ) {
       get_moddb_lock();
// First read the cache
       data = search_memhash(dbdtaq, chksum, TIGERLEN); 
       if (NULL == data) {
           LDEBUG("updateBlock : Not in dbdtaq");
           data = search_dbdata(dbdta, chksum, TIGERLEN);
           if ( NULL == data ) {
                 LDEBUG("updateBlock : Not in dbdta");
                 cachedata = search_memhash(blkcache, &inobno.inode, sizeof(unsigned long long));
                 if ( NULL != cachedata ) {
                   blk = (BLKCACHE *) cachedata->data;
                   dtiger=binhash(blk->blockdata, BLKSIZE);
                   if ( 0 == memcmp(dtiger,chksum,TIGERLEN)){
                      data=s_malloc(sizeof(DBT));
                      data->data=s_malloc(BLKSIZE);
                      data->size=BLKSIZE;
                      memcpy(data->data,blk->blockdata,BLKSIZE);
                      DBTfree(cachedata);
                   } else {
                     LDEBUG("updateBlock : Not in dbcache, out of luck.");
                     loghash("updateBlock : No data found to read ", chksum);
                     die_dataerr
                         ("No data found to read, this should never happen: inode :%llu: blocknr :%llu",
                          inode, blocknr);
                   }
                   free(dtiger);
                 } else {
                     loghash("updateBlock : No data found to read ", chksum);
                     die_dataerr
                         ("No data found to read, this should never happen: inode :%llu: blocknr :%llu",
                          inode, blocknr);
                 }
          }
       }
       release_moddb_lock();
       if (data->size < BLKSIZE) {
#ifdef LZO
           uncompdata = lzo_decompress(data->data, data->size);
#else
           LDEBUG("uncomp data");
           uncompdata = clz_decompress(data->data, data->size);
           LDEBUG("done uncomp data");
#endif
           LDEBUG("got uncompsize : %lu", uncompdata->size);
           memcpy(dbdata, uncompdata->data, uncompdata->size);
           comprfree(uncompdata);
       } else {
           LDEBUG("Got data->size %lu", data->size);
           memcpy(dbdata, data->data, data->size);
       }
       DBTfree(data);
    } 
    memcpy(dbdata + offsetblock, blockdata, size);
    add_blk_to_cache(inode, blocknr, dbdata);
    loghash("updateBlock -> commit_block : do delete hash", chksum);
    update_filesize(inode, size, offsetblock, blocknr, 0, 0,
                   0);
    free(dbdata);
    EFUNC;
    return;
}


int btdelete_curkey(TCBDB * db, void *key, int keylen, void *kvalue,
                    int kvallen)
{
    BDBCUR *cur;
    char *value;
    int vsize;
    int ksize;
    char *dbkey;
    int ret = 1;

    FUNC;
    cur = tcbdbcurnew(db);
    if (!tcbdbcurjump(cur, key, keylen)) {
        tcbdbcurdel(cur);
        return (-ENOENT);
    }
    /* traverse records */
    while ((dbkey = tcbdbcurkey(cur, &ksize)) != NULL) {
        if (0 != memcmp(dbkey, key, ksize))
            break;
        value = tcbdbcurval(cur, &vsize);
        if (value) {
            if (kvallen == vsize) {
                if (0 == memcmp(value, kvalue, kvallen)) {
                    ret = 0;
                    if (!tcbdbcurout(cur)) {
                        die_dataerr
                            ("Failed to delete key, this should never happen!");
                        ret = -ENOENT;
                    }
                    free(value);
                    free(dbkey);
                    break;
                }
            }
            free(value);
        }
        free(dbkey);
        tcbdbcurnext(cur);
    }
    tcbdbcurdel(cur);
    return (ret);
}

void *btsearch_keyval(TCBDB * db, void *key, int keylen, void *val,
                      int vallen)
{
    BDBCUR *cur;
    char *dbkey;
    char *dbvalue;
    void *ret = NULL;
    int size;

    FUNC;
    cur = tcbdbcurnew(db);
    if (!tcbdbcurjump(cur, key, keylen) && tcbdbecode(db) != TCESUCCESS) {
        tcbdbcurdel(cur);
        return ret;
    }
    /* traverse records */
    while ((dbkey = tcbdbcurkey(cur, &size)) != NULL) {
        if ( 0 != memcmp(dbkey,key,keylen)) break;
        dbvalue = tcbdbcurval(cur, &size);
        if (NULL != dbvalue) {
            if (NULL != val) {
                if (vallen == size) {
                    if (0 == memcmp(val, dbvalue, size)) {
                        ret = s_malloc(size + 1);
                        memset(ret, 0, size + 1);
                        memcpy(ret, dbvalue, size);
                        free(dbvalue);
                        free(dbkey);
                        break;
                    }
                }
            } else {
                ret = s_malloc(size + 1);
                memset(ret, 0, size + 1);
                memcpy(ret, dbvalue, size);
                free(dbvalue);
                free(dbkey);
                break;
            }
            free(dbvalue);
        }
        free(dbkey);
        tcbdbcurnext(cur);
    }
    tcbdbcurdel(cur);
    EFUNC;
    return ret;
}

int count_dirlinks(void *linkstr, int len)
{
    BDBCUR *cur;
    char *key;
    int size;
    int count = 0;

    cur = tcbdbcurnew(dbl);
    if (!tcbdbcurjump(cur, linkstr, len)
        && tcbdbecode(dbdirent) != TCESUCCESS) {
        tcbdbcurdel(cur);
        return (-ENOENT);
    }
    /* traverse records */
    while ((key = tcbdbcurkey(cur, &size)) != NULL) {
        if (len == size) {
            if (0 == memcmp(key, linkstr, len)) {
                count++;
            }
        }
        free(key);
        tcbdbcurnext(cur);
    }
    tcbdbcurdel(cur);
    return (count);
}

void delete_key(TCHDB * db, void *keydata, int len)
{
    int ecode;

    if (!tchdbout(db, keydata, len)) {
        ecode = tchdbecode(dbb);
        die_dataerr("Delete of key failed :%s", tchdberrmsg(ecode));
    }
}

void mdelete_key(TCMDB * db, void *keydata, int len)
{
    //FUNC;
    tcmdbout(db, keydata, len);
    //EFUNC;
}

DBT *check_block_exists(INOBNO inobno)
{
    DBT *data = NULL;
    DBT *cachedata = NULL;
    BLKCACHE *blk;

    FUNC;
    data = search_dbdata(dbb, &inobno, sizeof(INOBNO));
    if (NULL == data){
        cachedata = search_memhash(blkcache, &inobno.inode, sizeof(unsigned long long));
        if (cachedata != NULL) {
           blk = (BLKCACHE *) cachedata->data;
           if ((inobno.blocknr == blk->blocknr)) {
               data = s_malloc(sizeof(DBT));
               data->size = TIGERLEN;
               data->data = binhash(blk->blockdata,BLKSIZE);
           } else {
             LDEBUG("check_block_exists : %llu-%llu not found",inobno.inode,inobno.blocknr);
           }
           DBTfree(cachedata);
        } else {
             LDEBUG("check_block_exists : %llu-%llu not found",inobno.inode,inobno.blocknr);
             return NULL;
        }
    }
    return data;
}

void wait_io_pending(unsigned long long inode)
{
    int c;

    FUNC;
iowait:
    getLock();
    for (c = 0; c < max_threads; c++) {
        if (0 == tdta[c]->inode)
            continue;
        if (tdta[c]->inode == inode) {
            releaseLock();
            goto iowait;
        }
    }
    EFUNC;
    return;
}

int inode_block_pending(unsigned long long inode,
                        unsigned long long blocknr)
{
    int c;
    int pending = 0;

    FUNC;
    for (c = 0; c < max_threads; c++) {
        if (0 == tdta[c]->inode)
            continue;
        if (tdta[c]->inode == inode) {
            if (tdta[c]->blocknr == blocknr)
                pending = 1;
            break;
        }
    }
    EFUNC;
    return (pending);
}

int unlink_file(const char *path)
{
    int res = 0;
    int haslinks = 0;
    int dir_links = 0;
    struct stat st;
    struct stat dirst;
    char *dname;
    char *bname;
    unsigned char *stiger;
    unsigned long long inode;
    unsigned long long counter = 0;
    unsigned long long inuse;
    unsigned long long done = 0;
    time_t thetime;
    void *vdirnode;
    DBT *bdata;
    DBT *ddbuf;
    DBT *dataptr;
    DDSTAT *ddstat;
    DINOINO dinoino;
    INOBNO inobno;
    char *filename;

    FUNC;

    LDEBUG("unlink_file %s", path);
    res = dbstat(path, &st);
    if (res == -ENOENT)
        return (res);
    inode = st.st_ino;
    releaseLock();
    wait_io_pending(inode);
    flush_dta_queue(); 
    haslinks = st.st_nlink;
    thetime = time(NULL);
    dname = s_dirname((char *) path);
    /* Change ctime and mtime of the parentdir Posix std posix behavior */
    res = update_parent_time(dname);
    bname = s_basename((char *) path);
    res = dbstat(dname, &dirst);
    if (S_ISLNK(st.st_mode) && haslinks == 1 ) {
        LDEBUG("unlink symlink %s inode %llu", path, inode);
        delete_key(dbs, &inode, sizeof(unsigned long long));
        LDEBUG("unlink symlink done %s", path);
    }
    inobno.inode = inode;
    inobno.blocknr = counter;
    while (done < st.st_size) {
        bdata = search_dbdata(dbb, &inobno, sizeof(INOBNO));
        if (bdata == NULL) {
            LDEBUG("%llu is sparse", counter);
            done = done + BLKSIZE;
            counter++;
            inobno.blocknr = counter;
            continue;
        }
        stiger = s_malloc(bdata->size);
        memcpy(stiger, bdata->data, bdata->size);
        loghash("search inuse for ", stiger);
        inuse = getInUse(stiger);
        if ( inuse > 1 ) inuse--;
        LDEBUG("inuse=%llu",inuse);
        if (haslinks == 1) {
            if (1 == inuse) {
                delete_key(dbu, stiger, 24);
                delete_key(dbdta, stiger, 24);
            } else {
                loghash("updateInUse dbu,dbdta for ", stiger);
                updateInUse(stiger, 24, inuse);
            }
        }
        free(stiger);
        DBTfree(bdata);
        if (haslinks == 1) {
            delete_key(dbb, &inobno, sizeof(INOBNO));
        }
        counter++;
        inobno.blocknr = counter;
        done = done + BLKSIZE;
    }
    if (haslinks == 1) {
        if (0 !=
            (res =
            btdelete_curkey(dbdirent, &dirst.st_ino,
                            sizeof(unsigned long long), &inode,
                            sizeof(unsigned long long)))) {
            free(bname);
            free(dname);
            return (res);
        }
        delete_key(dbp, (unsigned char *) &inode,
                   sizeof(unsigned long long));
    } else {
        dataptr =
            search_dbdata(dbp, (unsigned char *) &inode,
                          sizeof(unsigned long long));
        if (dataptr == NULL) {
            die_dataerr("Failed to find file %llu", inode);
        }
        ddstat = value_to_ddstat((char *) dataptr->data, dataptr->size);
        ddstat->stbuf.st_nlink--;
        ddstat->stbuf.st_ctime = thetime;
        ddstat->stbuf.st_mtime = thetime;
        dinoino.dirnode = dirst.st_ino;
        dinoino.inode = ddstat->stbuf.st_ino;
        dir_links = count_dirlinks(&dinoino,sizeof(DINOINO));
        res =
            btdelete_curkey(dbl, &dinoino, sizeof(DINOINO), bname,
                            strlen(bname));
        btdelete_curkey(dbl, &ddstat->stbuf.st_ino,
                        sizeof(unsigned long long), &dinoino,
                        sizeof(DINOINO));
// Restore to regular file settings and clean up.
        if (ddstat->stbuf.st_nlink == 1) {
            vdirnode =
                btsearch_keyval(dbl, &ddstat->stbuf.st_ino,
                                sizeof(unsigned long long), NULL, 0);
            memcpy(&dinoino, vdirnode, sizeof(DINOINO));
            free(vdirnode);
            filename =
                (char *) btsearch_keyval(dbl, &dinoino, sizeof(DINOINO),
                                         NULL, 0);
            memcpy(&ddstat->filename, filename, strlen(filename) + 1);
            free(filename);
            LDEBUG
                ("unlink_file : Restore %s to regular file settings and clean up.",
                 ddstat->filename);
            btdelete_curkey(dbl, &dinoino, sizeof(DINOINO),
                            ddstat->filename, strlen(ddstat->filename));
            btdelete_curkey(dbl, &ddstat->stbuf.st_ino,
                            sizeof(unsigned long long), &dinoino,
                            sizeof(DINOINO));
            btdelete_curkey(dbl, &inode, sizeof(unsigned long long),
                            &dinoino, sizeof(DINOINO));
            res = 0;
        } 
        if (dir_links == 1 ){
            if (0 !=
                (res =
                 btdelete_curkey(dbdirent, &dirst.st_ino,
                                 sizeof(unsigned long long), &inode,
                                 sizeof(unsigned long long)))) {
                 die_dataerr("unlink_file : Failed to delete record.");
            }
        }
        ddbuf = create_ddbuf(ddstat->stbuf, ddstat->filename);
        bin_write_dbdata(dbp, &inode,
                         sizeof(unsigned long long), (void *) ddbuf->data,
                         ddbuf->size);
        DBTfree(dataptr);
        DBTfree(ddbuf);
        ddstatfree(ddstat);
    }
    free(bname);
    free(dname);
    EFUNC;
    return (res);
}

int fs_mkdir(const char *path, mode_t mode)
{
    unsigned long long inode;
    char *rdir;
    char *pdir;
    int rootdir = 0;
    int res;

    FUNC;
    LDEBUG("mode =%i", mode);
    if (0 == strcmp("/", path))
        rootdir = 1;
    inode = get_next_inode();
    write_file_ent(path, inode, S_IFDIR | mode, NULL, 0);
    if (rootdir) {
        rdir = as_sprintf("%s.", path);
    } else {
        rdir = as_sprintf("%s/.", path);
    }
    inode = get_next_inode();
    write_file_ent(rdir, inode, S_IFDIR | 0755, NULL, 0);
    free(rdir);
    if (rootdir) {
        rdir = as_sprintf("%s..", path);
    } else {
        rdir = as_sprintf("%s/..", path);
    }
    inode = get_next_inode();
    write_file_ent(rdir, inode, S_IFDIR | 0755, NULL, 0);
    free(rdir);
    /* Change ctime and mtime of the parentdir Posix std posix behavior */
    pdir = s_dirname((char *) path);
    res = update_parent_time(pdir);
    free(pdir);
    return (res);
}

void partial_truncate_block(struct stat *stbuf, unsigned long long blocknr,unsigned int offset)
{
    unsigned char *blockdata;
    compr *uncompdata;
    INOBNO inobno;
    DBT *data;
    unsigned char *stiger;
    unsigned long long inuse;


    FUNC;
    LDEBUG("partial_truncate_block : inode %llu, blocknr %llu, offset %u",stbuf->st_ino,blocknr,offset);
    inobno.inode = stbuf->st_ino;
    inobno.blocknr = blocknr;
    data = search_dbdata(dbb, &inobno, sizeof(INOBNO));
    if (NULL == data) {
        LDEBUG("Deletion of non existent block.");
        return;
    }
    stiger = s_malloc(data->size);
    loghash("partial_truncate_block : search tiger ",stiger); 
    memcpy(stiger, data->data, data->size);
    DBTfree(data);
    data=search_dbdata(dbdta, stiger, TIGERLEN);
    inuse = getInUse(stiger);
    if (inuse == 1) {
        loghash("partial_truncate_block : delete hash",stiger);
        delete_key(dbdta, stiger, TIGERLEN);
        delete_key(dbu, stiger, TIGERLEN);
        delete_key(dbb, &inobno, sizeof(INOBNO));
    } else {
        if (inuse > 1 ) inuse--;
        delete_key(dbb, &inobno, sizeof(INOBNO));
        updateInUse(stiger, TIGERLEN, inuse);
    }
    blockdata=s_malloc(BLKSIZE);
    memset(blockdata,0,BLKSIZE);
    if (data->size != BLKSIZE) {
#ifdef LZO
        uncompdata = lzo_decompress(data->data, data->size);
#else
        uncompdata = clz_decompress(data->data, data->size);
#endif
        memcpy(blockdata, uncompdata->data, offset);
        comprfree(uncompdata);
    } else {
        memcpy(blockdata, data->data, offset);
    }
    add_blk_to_cache(inobno.inode, inobno.blocknr, blockdata);
    free(stiger);
    DBTfree(data);
    free(blockdata); 
    return;
}

int fs_truncate(struct stat *stbuf, off_t size, char *bname)
{
    unsigned int offsetblock;
    unsigned long long blocknr;
    unsigned long long lastblocknr;
    unsigned long long inuse;
    unsigned char *stiger;
    off_t oldsize;
    DBT *data;
    MEMDDSTAT *memddstat;
    DDSTAT *ddstat;
    DBT *ddbuf;
    INOBNO inobno;
    time_t thetime;

    FUNC;
    LDEBUG("lessfs_truncate inode %llu - size %llu", stbuf->st_ino,
           (unsigned long long) size);
    thetime = time(NULL);
    blocknr = size / BLKSIZE;
    offsetblock = size - (blocknr * BLKSIZE);
    oldsize = stbuf->st_size;
    lastblocknr = oldsize / BLKSIZE;
    // Truncate filesize.
    data = search_dbdata(dbp, &stbuf->st_ino, sizeof(unsigned long long));
    if (NULL == data) {
        LDEBUG("fs_truncate : no inode found to truncate");
        return (-ENOENT);
    }
    ddstat = value_to_ddstat((char *) data->data, data->size);
    DBTfree(data);
    data = search_memhash(dbcache, &stbuf->st_ino,
                           sizeof(unsigned long long));
    if ( NULL != data ) {
        memddstat = value_tomem_ddstat((char *) data->data, data->size);
        memcpy(&memddstat->stbuf, stbuf, sizeof(struct stat));
        memddstat->stbuf.st_size = size;
        memddstat->stbuf.st_ctime = thetime;
        memddstat->stbuf.st_mtime = thetime;
        memddstat->updated = 1;
        ddbuf = create_mem_ddbuf(memddstat);
        mbin_write_dbdata(dbcache, &stbuf->st_ino, sizeof(unsigned long long),
                      (void *) ddbuf->data, ddbuf->size);
        DBTfree(ddbuf);
        memddstatfree(memddstat);
        DBTfree(data);
    } else { 
            data = search_dbdata(dbp, &stbuf->st_ino, sizeof(unsigned long long));
            if (NULL == data) {
                return (-ENOENT);
            }
            ddstat = value_to_ddstat((char *) data->data, data->size);
            ddstat->stbuf.st_mtime = thetime;
            ddstat->stbuf.st_ctime = thetime;
            ddstat->stbuf.st_size = size;
            DBTfree(data);
            data = create_ddbuf(ddstat->stbuf, ddstat->filename);
            bin_write_dbdata(dbp, &stbuf->st_ino, sizeof(unsigned long long),
                     (void *) data->data, data->size);
            DBTfree(data);
    }
    ddstatfree(ddstat);
    LDEBUG("lessfs_truncate : truncate new block %llu, oldblock %llu",blocknr,lastblocknr);
    while (lastblocknr > blocknr) {
        LDEBUG("lessfs_truncate : Enter loop lastblocknr %llu : blocknr %llu",lastblocknr,blocknr);
        inobno.inode = stbuf->st_ino;
        inobno.blocknr = lastblocknr;
        data = search_dbdata(dbb, &inobno, sizeof(INOBNO));
        if (NULL == data) {
            LDEBUG("Deletion of non existent block inode : %llu, blocknr %llu",inobno.inode,inobno.blocknr);
            if ( lastblocknr > 0 ) 
               lastblocknr--;
            else break;
// Need to continue in case of a sparse file.
            continue;
        }
        stiger = s_malloc(data->size);
        memcpy(stiger, data->data, data->size);
        LDEBUG("lessfs_truncate Search to delete blocknr %llu:", lastblocknr); 
        loghash("lessfs_truncate tiger :",stiger);
        DBTfree(data);
        inuse = getInUse(stiger);
        if (inuse == 1) {
            flush_dta_queue();
            loghash("truncate : delete hash",stiger);
            delete_key(dbdta, stiger, TIGERLEN);
            delete_key(dbu, stiger, TIGERLEN);
            delete_key(dbb, &inobno, sizeof(INOBNO));
        } else {
            if ( inuse > 1 ) inuse--;
            delete_key(dbb, &inobno, sizeof(INOBNO));
            updateInUse(stiger, TIGERLEN, inuse);
        }
        if ( lastblocknr > 0 ) lastblocknr--;
        free(stiger);
    }
    LDEBUG("offsetblock = %u",offsetblock);
    if ( 0 != offsetblock ) partial_truncate_block(stbuf,lastblocknr,offsetblock);
    return (0);
}

int update_parent_time(char *path)
{
    int res;
    struct stat stbuf;
    time_t thetime;

    FUNC;
    LDEBUG("update_parent_time : %s", path);
    thetime = time(NULL);
    /* Change ctime and mtime of the parentdir Posix std posix behavior */
    res = dbstat(path, &stbuf);
    if (0 != res)
        return (res);
    stbuf.st_ctime = thetime;
    stbuf.st_mtime = thetime;
    res = update_stat(path, &stbuf);
    EFUNC;
    return (res);
}

int fs_rmdir(const char *path)
{
    int res;
    char *dotstr;
    char *dotdotstr;
    char *dname;
    unsigned long long dirnode;
    unsigned long long keynode;
    unsigned long long pathnode;
    unsigned long long dirnodes;

    FUNC;
    LDEBUG("rmdir called : %s", path);

    pathnode = get_inode(path);
    dirnodes = has_nodes(pathnode);
    if (0 != dirnodes) {
        LDEBUG("fs_rmdir : Cannot remove directory %s with %llu files",
               path, dirnodes);
        EFUNC;
        return -ENOTEMPTY;
    }

    dname = s_dirname((char *) path);
    dotstr = as_sprintf("%s/.", path);
    dotdotstr = as_sprintf("%s/..", path);

    keynode = get_inode(dotstr);
    LDEBUG("inode for %s is %llu", dotstr, keynode);
    delete_key(dbp, &keynode, sizeof(unsigned long long));
    btdelete_curkey(dbdirent, &pathnode, sizeof(unsigned long long),
                    &keynode, sizeof(unsigned long long));

    keynode = get_inode(dotdotstr);
    delete_key(dbp, &keynode, sizeof(unsigned long long));
    btdelete_curkey(dbdirent, &pathnode, sizeof(unsigned long long),
                    &keynode, sizeof(unsigned long long));

    dirnode = get_inode(dname);
    delete_key(dbp, &pathnode, sizeof(unsigned long long));
    btdelete_curkey(dbdirent, &dirnode, sizeof(unsigned long long),
                    &pathnode, sizeof(unsigned long long));
    btdelete_curkey(dbdirent, &pathnode, sizeof(unsigned long long),
                    &pathnode, sizeof(unsigned long long));
    free(dotstr);
    free(dotdotstr);
    res = update_parent_time(dname);
    free(dname);
    return (res);
}

unsigned long long get_inode(const char *path)
{
    struct stat stbuf;

    FUNC;
    if (0 != dbstat(path, &stbuf)) {
        LDEBUG("get_inode : nothing found for %s", path);
        return (0);
    }
    EFUNC;
    return (stbuf.st_ino);
}

unsigned long long has_nodes(unsigned long long inode)
{
    unsigned long long res = 0;
    BDBCUR *cur;
    unsigned long long filenode;
    char *filenodestr;
    DBT *filedata;
    bool dotdir = 0;
    char *key;
    DDSTAT *ddstat;
    int size;
    unsigned long long keyval;

    FUNC;
    cur = tcbdbcurnew(dbdirent);
    if (!tcbdbcurjump(cur, &inode, sizeof(unsigned long long))
        && tcbdbecode(dbdirent) != TCESUCCESS) {
        tcbdbcurdel(cur);
        return (-ENOENT);
    }
    while ((key = tcbdbcurkey(cur, &size)) != NULL) {
        memcpy(&keyval, key, sizeof(unsigned long long));
        if (keyval != inode) {
            free(key);
            break;
        }
        filenodestr = (char *) tcbdbcurval(cur, &size);
        memcpy(&filenode, filenodestr, sizeof(unsigned long long));
        filedata =
            search_dbdata(dbp, &filenode, sizeof(unsigned long long));
        if (NULL != filedata) {
            ddstat =
                value_to_ddstat((char *) filedata->data, filedata->size);
            DBTfree(filedata);
            LDEBUG("Compare : %llu %llu", filenode, inode);
            if (filenode == inode)
                dotdir = 1;
            if (NULL != ddstat->filename) {
                if (0 == strcmp(ddstat->filename, "."))
                    dotdir = 1;
                if (0 == strcmp(ddstat->filename, ".."))
                    dotdir = 1;
            }
            if (!dotdir) {
                LDEBUG
                    ("has_nodes : Found file in directory %s filenode %llu inode %llu",
                     ddstat->filename, filenode, inode);
                res++;
            }
            ddstatfree(ddstat);
        }
        free(key);
        free(filenodestr);
        tcbdbcurnext(cur);
        dotdir = 0;
    }
    tcbdbcurdel(cur);
    LDEBUG("inode %llu contains %llu files", inode, res);
    EFUNC;
    return (res);
}

void fs_read_hardlink(struct stat stbuf, DDSTAT * ddstat, void *buf,
                      fuse_fill_dir_t filler, struct fuse_file_info *fi)
{
    BDBCUR *cur;
    char *linkkey;
    int ksize;
    DINOINO dinoino;
    char *filename;

    FUNC;
    cur = tcbdbcurnew(dbl);
    dinoino.dirnode = stbuf.st_ino;
    dinoino.inode = ddstat->stbuf.st_ino;
    if (!tcbdbcurjump(cur, &dinoino, sizeof(DINOINO))
        && tcbdbecode(dbl) != TCESUCCESS)
        die_dataerr("Unable to find linkname, run fsck.");
    while ((linkkey = tcbdbcurkey(cur, &ksize)) != NULL) {
        if (0 != memcmp(linkkey, &dinoino, sizeof(DINOINO))) {
            LDEBUG("fs_read_hardlink : linkkey != dinoino");
            free(linkkey);
            break;
        }
        filename = (char *) tcbdbcurval(cur, &ksize);
        memcpy(&ddstat->filename, filename, strlen(filename) + 1);
        free(filename);
        LDEBUG("fs_read_hardlink : fil_fuse_info %s size %i",
               ddstat->filename, ksize);
        fil_fuse_info(ddstat, buf, filler, fi);
        ddstat->filename[0] = 0;
        free(linkkey);
        tcbdbcurnext(cur);
    }
    tcbdbcurdel(cur);
    EFUNC;
    return;
}

void fil_fuse_info(DDSTAT * ddstat, void *buf, fuse_fill_dir_t filler,
                   struct fuse_file_info *fi)
{
    struct stat st;
    char *bname;

    memcpy(&st, &ddstat->stbuf, sizeof(struct stat));
    bname = s_basename(ddstat->filename);
    if (NULL != bname) {
        // Don't show the directory
        if (0 != strcmp(bname, "/")) {
            LDEBUG("Fill filler with %s", bname);
            filler(buf, bname, &st, 0);
        }
    }
    free(bname);
}

int fs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
               off_t offset, struct fuse_file_info *fi)
{
    int retcode = 0;
    BDBCUR *cur;
    int size, res;
    int ksize;
    char *key;
    struct stat stbuf;
    unsigned long long keynode;
    unsigned long long filenode;
    char *filenodestr;
    DBT *filedata;
    DDSTAT *ddstat;
    FUNC;

    (void) offset;
    (void) fi;
    LDEBUG("Called fs_readdir with path %s", (char *) path);

    res = dbstat(path, &stbuf);
    if (0 != res)
        return -ENOENT;
    cur = tcbdbcurnew(dbdirent);
    if (!tcbdbcurjump(cur, &stbuf.st_ino, sizeof(unsigned long long))
        && tcbdbecode(dbdirent) != TCESUCCESS) {
        tcbdbcurdel(cur);
        return (-ENOENT);
    }
    while ((key = tcbdbcurkey(cur, &ksize)) != NULL) {
        memcpy(&keynode, key, sizeof(unsigned long long));
        if (stbuf.st_ino != keynode) {
            free(key);
            break;
        }
        filenodestr = tcbdbcurval(cur, &size);
        memcpy(&filenode, filenodestr, sizeof(unsigned long long));
        if (filenode == keynode) {
            free(filenodestr);
            free(key);
            tcbdbcurnext(cur);
            continue;
        }
        LDEBUG("GOT filenode %llu", filenode);
        filedata =
            search_dbdata(dbp, &filenode, sizeof(unsigned long long));
        if (NULL != filedata) {
            ddstat =
                value_to_ddstat((char *) filedata->data, filedata->size);
            DBTfree(filedata);
            if (ddstat->filename[0] == 0) {
                fs_read_hardlink(stbuf, ddstat, buf, filler, fi);
            } else {
                fil_fuse_info(ddstat, buf, filler, fi);
            }
            ddstatfree(ddstat);
        }
        free(key);
        free(filenodestr);
        tcbdbcurnext(cur);
    }
    tcbdbcurdel(cur);
    LDEBUG("fs_readdir: return");
    return (retcode);
}

int fs_link(char *from, char *to)
{
    int res = 0;
    unsigned long long inode;
    unsigned long long todirnode;
    unsigned long long fromdirnode;
    struct stat stbuf;
    struct stat tobuf;
    struct stat frombuf;
    char *bfrom;
    char *bto;
    char *todir;
    char *fromdir;
    DBT *ddbuf;
    DBT *symdata;
    time_t thetime;
    DINOINO dinoino;

    FUNC;
    LDEBUG("fs_link called with from=%s to=%s", from, to);
    res = dbstat(from, &stbuf);

    if (res != 0)
        return (res);
    fromdir = s_dirname(from);
    res = dbstat(fromdir, &frombuf);
    todir = s_dirname(to);
    res = dbstat(todir, &tobuf);
    if (res != 0) {
        free(todir);
        free(fromdir);
        return -ENOENT;
    }
    bfrom = s_basename(from);
    bto = s_basename(to);

/* Update inode nrlinks */
    todirnode = tobuf.st_ino;
    fromdirnode = frombuf.st_ino;
    inode = stbuf.st_ino;
    // Update nr of links.
    stbuf.st_nlink++;
    thetime = time(NULL);
    stbuf.st_ctime = thetime;
    update_cache(inode, &stbuf);
    ddbuf = create_ddbuf(stbuf, NULL);
    LDEBUG("fs_link : update links on %llu to %i", inode, stbuf.st_nlink);
    bin_write_dbdata(dbp, &inode, sizeof(unsigned long long),
                     ddbuf->data, ddbuf->size);
    DBTfree(ddbuf);

/* Link dest filename inode to dest directory if it does not exist*/
    if (0 ==
        bt_entry_exists(dbdirent, &todirnode, sizeof(unsigned long long),
                        &inode, sizeof(unsigned long long))) {
        LDEBUG("fs_link : write link %llu : %llu", todirnode, inode);
        btbin_write_dup(dbdirent, &todirnode, sizeof(unsigned long long),
                        &inode, sizeof(unsigned long long));
    }

/* Link some more, hardlink a symlink. */
    if (S_ISLNK(stbuf.st_mode)) {
        LFATAL("fs_link : hardlink a symlink");
        symdata =
            search_dbdata(dbs, &inode, sizeof(unsigned long long));
        if (NULL == symdata)
            die_dataerr("Unable to read symlink");
        DBTfree(symdata);
    }
/* Write L_destinodedir_inode : dest filename */
    dinoino.dirnode = tobuf.st_ino;
    dinoino.inode = stbuf.st_ino;
    LDEBUG("A. fs_link : write link %llu-%llu : %s", tobuf.st_ino,
           stbuf.st_ino, bto);
    btbin_write_dup(dbl, &dinoino, sizeof(DINOINO), bto, strlen(bto));
    btbin_write_dup(dbl, &stbuf.st_ino, sizeof(unsigned long long),
                    &dinoino, sizeof(DINOINO));

/* Write Lfrominode_inode : from filename */
    dinoino.dirnode = frombuf.st_ino;
    dinoino.inode = stbuf.st_ino;
    if ( stbuf.st_nlink == 2 ) {
       btbin_write_dup(dbl, &dinoino, sizeof(DINOINO), bfrom,
                       strlen(bfrom));
       btbin_write_dup(dbl, &stbuf.st_ino, sizeof(unsigned long long),
                       &dinoino, sizeof(DINOINO));
    }
    res = update_parent_time(todir);
    free(todir);
    free(fromdir);
    free(bfrom);
    free(bto);
    EFUNC;
    return (res);
}

int bt_entry_exists(TCBDB * db, void *parent, int parentlen, void *value,
                    int vallen)
{
    int res = 0;
    BDBCUR *cur;
    char *dbvalue;
    char *key;
    int ksize;

    FUNC;
    cur = tcbdbcurnew(dbdirent);
    if (!tcbdbcurjump(cur, parent, parentlen)
        && tcbdbecode(db) != TCESUCCESS) {
        tcbdbcurdel(cur);
        return (res);
    }
    while ((key = tcbdbcurkey(cur, &ksize)) != NULL) {
        if (ksize != parentlen) {
            free(key);
            break;
        }
        if (0 != memcmp(key, parent, parentlen)) {
            free(key);
            break;
        }
        dbvalue = tcbdbcurval2(cur);
        if (dbvalue) {
            if (0 == memcmp(value, dbvalue, vallen))
                res = 1;
            free(dbvalue);
        }
        free(key);
        tcbdbcurnext(cur);
    }
    tcbdbcurdel(cur);
    LDEBUG("bt_entry_exists : returns %i", res);
    EFUNC;
    return (res);
}

int fs_symlink(char *from, char *to)
{
    int res = 0;
    char *todir;

    todir = s_dirname(to);
    dbmknod(to, 0777 | S_IFLNK, from, 0);
    res = update_parent_time(todir);
    free(todir);
    return (res);
}

int fs_readlink(const char *path, char *buf, size_t size)
{
    int res = 0;
    DBT *data;
    unsigned long long inode;

    FUNC;
    inode = get_inode(path);
    if (0 == inode)
        return (-ENOENT);

    data = search_dbdata(dbs, &inode, sizeof(unsigned long long));
    if (NULL == data) {
        res = -ENOENT;
    } else {
        if (size - 1 > data->size) {
            memcpy(buf, data->data, data->size + 1);
        } else {
            memcpy(buf, data->data, size - 1);
        }
        DBTfree(data);
    }
    return (res);
}

int fs_rename_link(const char *from, const char *to, struct stat stbuf)
{
    unsigned long long fromnode;
    unsigned long long tonode;
    unsigned long long inode;
    char *fromdir;
    char *todir;
    char *bfrom;
    char *bto;
    struct stat st;
    int res = 0;
    DINOINO dinoino;

    FUNC;

    LDEBUG("fs_rename_link from: %s : to %s", (char *) from, (char *) to);
    if (0 == strcmp(from, to))
        return (0);
    if (-ENOENT != dbstat(to, &st))
        unlink_file(to);
    fromdir = s_dirname((char *) from);
    todir = s_dirname((char *) to);
    bfrom = s_basename((char *) from);
    bto = s_basename((char *) to);
    inode = stbuf.st_ino;

    fromnode = get_inode(fromdir);
    tonode = get_inode(todir);
    if (0 == fromnode)
        die_dataerr("Unable to find directory %s for file %s", fromdir,
                    from);
    if (0 == tonode)
        die_dataerr("Unable to find directory %s for file %s", todir, to);

    dinoino.dirnode = fromnode;
    dinoino.inode = stbuf.st_ino;
    btdelete_curkey(dbl, &dinoino, sizeof(DINOINO), bfrom, strlen(bfrom));
    if (count_dirlinks(&dinoino, sizeof(DINOINO)) > 1) {
        btdelete_curkey(dbdirent, &fromnode, sizeof(unsigned long long),
                        &inode, sizeof(unsigned long long));
        btbin_write_dup(dbdirent, &tonode, sizeof(unsigned long long),
                        &inode, sizeof(unsigned long long));
    }
    dinoino.dirnode = tonode;
    dinoino.inode = stbuf.st_ino;
    btbin_write_dup(dbl, &dinoino, sizeof(DINOINO), bto, strlen(bto));
    free(fromdir);
    free(bfrom);
    free(bto);
    free(todir);
    EFUNC;
    return (res);
}

void update_cache(unsigned long long inode, struct stat *stbuf)
{
    DBT *dataptr;
    MEMDDSTAT *ddstat;
    DBT *ddbuf;

    FUNC;
    LDEBUG("update_cache nlinks : %u", stbuf->st_nlink);
    dataptr = search_memhash(dbcache, &inode, sizeof(unsigned long long));
    if (dataptr == NULL) {
        return;
    }
    ddstat = value_tomem_ddstat((char *) dataptr->data, dataptr->size);
    memcpy(&ddstat->stbuf, &stbuf, sizeof(struct stat));
    ddbuf = create_mem_ddbuf(ddstat);
    mbin_write_dbdata(dbcache, &inode, sizeof(unsigned long long),
                      (void *) ddbuf->data, ddbuf->size);
    DBTfree(dataptr);
    DBTfree(ddbuf);
    hash_update_filesize(ddstat, inode);
    ddstat->updated = 0;
    memddstatfree(ddstat);
    EFUNC;
    return;
}

int fs_rename(const char *from, const char *to, struct stat stbuf)
{
    DBT *dataptr;
    DDSTAT *ddstat;
    DBT *ddbuf;
    int res = 0;
    unsigned long long inode;
    time_t thetime;
    char *bto;
    char *bfrom;
    char *fromdir;
    char *todir;
    unsigned long long fromdirnode;
    unsigned long long todirnode;
    unsigned long long tonode;
    struct stat st;
    unsigned long long dirnodes;

    FUNC;

    LDEBUG("fs_rename from: %s : to %s", (char *) from, (char *) to);
    LDEBUG("fs_rename : stat %s", to);
    todir = s_dirname((char *) to);
    todirnode = get_inode(todir);
    tonode = get_inode(to);
    bto = s_basename((char *) to);
    bfrom = s_basename((char *) from);
    if (-ENOENT != dbstat(to, &st)) {
        if (S_ISDIR(st.st_mode)) {
            LDEBUG("fs_rename bto %s bfrom %s", bto, bfrom);
            dirnodes = has_nodes(tonode);
            if (0 != dirnodes) {
                if (0 != strcmp(bto, bfrom)) {
                    LDEBUG
                        ("fs_rename : Cannot rename directory %llu with %llu files",
                         todirnode, dirnodes);
                    free(todir);
                    free(bfrom);
                    free(bto);
                    EFUNC;
                    return -ENOTEMPTY;
                }
            }
        }
        LDEBUG("fs_rename : destination file %s exists, unlink_file.", to);
        unlink_file(to);
    }
    free(bfrom);
    inode = stbuf.st_ino;
    fromdir = s_dirname((char *) from);
    fromdirnode = get_inode(fromdir);
    LDEBUG("fs_rename : bto = %s", bto);
    dataptr = search_dbdata(dbp, &inode, sizeof(unsigned long long));
    if (dataptr == NULL) {
        die_dataerr("Failed to find file %llu", inode);
    }
    ddstat = value_to_ddstat((char *) dataptr->data, dataptr->size);
    thetime = time(NULL);
    ddstat->stbuf.st_ctime = thetime;
    ddbuf = create_ddbuf(ddstat->stbuf, (char *) bto);
    bin_write_dbdata(dbp, &inode,
                     sizeof(unsigned long long), (void *) ddbuf->data,
                     ddbuf->size);
    if (fromdirnode != todirnode) {
        LDEBUG("fs_rename : rename inode %llu : %llu to another path %llu",
               inode, fromdirnode, todirnode);
        btdelete_curkey(dbdirent, &fromdirnode, sizeof(unsigned long long),
                        &inode, sizeof(unsigned long long));
        btbin_write_dup(dbdirent, &todirnode, sizeof(unsigned long long),
                        &inode, sizeof(unsigned long long));
    }
    DBTfree(dataptr);
    DBTfree(ddbuf);
    ddstatfree(ddstat);
    free(bto);
    free(fromdir);
    free(todir);
    EFUNC;
    return (res);
}

int update_stat(char *path, struct stat *stbuf)
{
    DDSTAT *ddstat;
    DBT *ddbuf;
    DBT *dataptr;
    unsigned long long inode;

    FUNC;
    inode = stbuf->st_ino;
    dataptr = search_dbdata(dbp, &inode, sizeof(unsigned long long));
    if (dataptr == NULL) {
        return (-ENOENT);
    }
    ddstat = value_to_ddstat((char *) dataptr->data, dataptr->size);
    memcpy(&ddstat->stbuf, stbuf, sizeof(struct stat));
    ddbuf = create_ddbuf(ddstat->stbuf, ddstat->filename);
    bin_write_dbdata(dbp, &inode,
                     sizeof(unsigned long long), (void *) ddbuf->data,
                     ddbuf->size);
    DBTfree(dataptr);
    DBTfree(ddbuf);
    ddstatfree(ddstat);
    EFUNC;
    return (0);
}
